 -- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use work.sha3_pkg.all;
use work.groestl_pkg.all;

-- Groestl datapath for quasi-pipelined architecture
-- possible generics values: HS = {GROESTL_DATA_SIZE_SMALL, GROESTL_DATA_SIZE_BIG}
-- ROMSTYLE = {DISTRIBUTED, COMBINATIONAL}
-- all combinations are allowed

entity groestl_datapath_pq_parallel_ppl is
generic (   n	: integer := GROESTL_DATA_SIZE_SMALL;
            hs 	: integer := HASH_SIZE_256;	  
			ppl : integer := 2;
            rom_style : integer := DISTRIBUTED
         );
port(
	-- external
	rst     : in std_logic;
	clk     : in std_logic;
	din     : in std_logic_vector(ppl*w-1 downto 0);	
	dout    : out std_logic_vector(w-1 downto 0);	
	

	-- input
	en_len				:in std_logic_vector(ppl-1 downto 0);	
	en_ctr				:in std_logic_vector(ppl-1 downto 0);	
	ein 				:in std_logic_vector(ppl-1 downto 0);	
	c					:out std_logic_vector(ppl*16-1 downto 0);	

	-- processing		  
	smsg        		: in std_logic_vector(log2(ppl)-1 downto 0);
	schain				: in std_logic_vector(log2(ppl)-1 downto 0);
	init1				: in std_logic;
	init2				: in std_logic;
	init3				: in std_logic;
	init4				: in std_logic;
	finalization		: in std_logic;
	wr_state			: in std_logic;
	wr_result			: in std_logic_vector(ppl-1 downto 0);
	load_ctr			: in std_logic;
	wr_ctr				: in std_logic;

	-- Output				
	shift_dout  : in std_logic_vector(ppl-1 downto 0);
    load_id     : in std_logic_vector(ppl-1 downto 0);
    load_hash   : in std_logic_vector(ppl-1 downto 0);
	sout        : in std_logic_vector(log2(ppl)-1 downto 0);
	sid         : in std_logic	
	);
end groestl_datapath_pq_parallel_ppl;

architecture basic of groestl_datapath_pq_parallel_ppl is	 
	-- type declaration																 
	type block_array_type is array (0 to ppl-1) of std_logic_vector(n-1 downto 0);	
	signal message : block_array_type;
	signal from_final_reg : block_array_type;
	
	--
	constant log2mw : integer := log2( n );
	constant log2mwzeroes : std_logic_vector(log2mw-1 downto 0) := (others => '0');
	constant zero : std_logic_vector(n-1 downto 0):= (others=>'0');
    signal rin, from_final, to_left_round, to_right_round, from_left_register, from_right_register, to_final : std_logic_vector(n-1 downto 0);
    signal to_left_register, to_right_register, intermediate_left, two_xor, three_xor, to_init : std_logic_vector(n-1 downto 0);
	signal to_rin_xor : std_logic_vector(n-1 downto 0);
	signal ctr : std_logic_vector(3 downto 0);
	signal round : std_logic_vector(7 downto 0);
	
	signal from_final2 : std_logic_vector(n-1 downto 0);
	
	-- id, hash		   
	constant id_size : integer := 16;
	signal id_in : std_logic_vector(ppl*id_size-1 downto 0);
	
begin
	-- ///////////////////////////////////////////////////////
	-- FSM1 SECTION
	fsm1_section : for i in 0 to ppl-1 generate
		decounter_gen : decountern 
			generic map ( N => 16, sub => 1 ) 
			port map ( clk => clk, rst => '0', load => en_len(i), en => en_ctr(i), 
						input => din(ppl*w-(i+1)*w+log2mw+15 downto ppl*w-(i+1)*w+log2mw), 
						output => c(PPL*16-(i+1)*16+15 downto PPL*16-(i+1)*16));
																																									
		shfin_gen : sipo 
			generic map ( N => n,M => w) 
			port map (clk => clk, en => ein(i), input => din(ppl*w-i*w-1 downto ppl*w-(i+1)*w), output => message(i) );
	end generate;
	
	-- ///////////////////////////////////////////////////////
	-- FSM2 SECTION
	-- Message selectors	 
	ppl2 : if ppl = 2 generate
		rin <= message(0) when smsg(0) = '0' else message(1);
																		   
		-- chain value for xor with the message
		from_final <= from_final_reg(0) when smsg(0) = '0' else from_final_reg(1);	
							
		-- chain value for xor with the round
		from_final2 <= from_final_reg(0) when schain(0) = '0' else from_final_reg(1);
	end generate;	   
	
	ppl4 : if ppl = 4 generate   
		with smsg(1 downto 0) select		
		rin <= message(0) when "00", message(1) when "01", message(2) when "10", message(3) when others;		
		
		with smsg(1 downto 0) select
		from_final <= 	from_final_reg(0) when "00",
						from_final_reg(1) when "01",
						from_final_reg(2) when "10",
						from_final_reg(3) when others;	 
						
		with schain(1 downto 0) select
		from_final2 <= 	from_final_reg(0) when "00",
						from_final_reg(1) when "01",
						from_final_reg(2) when "10",
						from_final_reg(3) when others;
	end generate;	  
	
	ChainValueGen: for i in 0 to ppl-1 generate
		ChainValueReg:  regn
			generic map(N=>n, init=>zero(n-1 downto 0))
			port map (clk => clk, rst => rst, en => wr_result(i), input => to_final, output => from_final_reg(i) );
	end generate;
	

	intermediate_left <= from_final when init2='1' else from_left_register;
	to_left_round <= 	(rin xor to_rin_xor) when init1='1' else intermediate_left;
	to_right_round <= 	rin  when init1='1' else from_right_register;

	-- round counter
	rd_num : countern
	generic map (N =>4, step=>1, style =>COUNTER_STYLE_1)
	port map (clk=>clk, rst=>rst, load=>load_ctr, en=>wr_ctr, input=> zero(3 downto 0) ,  output=>ctr);
	round <= zero(3 downto 0) & ctr;  
	
	-- parallel round
	left_round : entity work.groestl_pq_parallel_ppl(struct)
		generic map (n=>n, rom_style => rom_style, p_mode => 1, ppl=>ppl)
		port map (clk=>clk, round=>round, input=>to_left_round, output=>to_left_register);

	right_round : entity work.groestl_pq_parallel_ppl(struct)
		generic map (n=>n, rom_style => rom_style, p_mode => 0, ppl=>ppl)
		port map (clk=>clk, round=>round, input=>to_right_round, output=>to_right_register);

	-- storage register for intermediate values
	lstate_reg : regn
	generic map(N=>n, init=>zero(n-1 downto 0))
	port map (clk => clk, rst => rst, en => wr_state, input => to_left_register, output => from_left_register );

	rstate_reg : regn
	generic map(N=>n, init=>zero(n-1 downto 0))
	port map (clk => clk, rst => rst, en => wr_state, input => to_right_register, output => from_right_register );

	two_xor <= from_final2 xor to_left_register;
	three_xor <= two_xor xor to_right_register;

	to_init <= three_xor;

	-- initialization vectors for different versions of Groestl
	iv224: if hs=HASH_SIZE_224 generate
		to_final <= GROESTL_INIT_VALUE_224 when init3='1' else to_init;
		to_rin_xor   <= GROESTL_INIT_VALUE_224 when init4='1' else from_final;
	end generate;

	iv256: if hs=HASH_SIZE_256 generate
		to_final <= GROESTL_INIT_VALUE_256 when init3='1' else to_init;
		to_rin_xor   <= GROESTL_INIT_VALUE_256 when init4='1' else from_final;
	end generate;

	iv384: if hs=HASH_SIZE_384 generate
		to_final <= GROESTL_INIT_VALUE_384 when init3='1' else to_init;	 
		to_rin_xor   <= GROESTL_INIT_VALUE_384 when init4='1' else from_final;
	end generate;

	iv512: if hs=HASH_SIZE_512 generate
		to_final <= GROESTL_INIT_VALUE_512 when init3='1' else to_init;
		to_rin_xor   <= GROESTL_INIT_VALUE_512 when init4='1' else from_final;
	end generate;
	
 	-- ///////////////////////////////////////////////////////
	-- FSM3 SECTION
	-- Round Finalization / Load new block / Output generation
	fsm3_output : entity work.ppl_output_unit(struct) 
		generic map ( h => HS, w => w, ppl => ppl ) 
		port map ( clk => clk, rst => rst,
					hash_result => two_xor(HS-1 downto 0), id_in => id_in,
					sid => sid, shift_dout => shift_dout, load_hash => load_hash, 
					load_id => load_id, sout => sout,
					dout => dout );			 
			
	id_in_gen : for i in 0 to ppl-1 generate
		id_in((ppl-i)*id_size-1 downto (ppl-i-1)*id_size) <= din(ppl*w-(i+1)*w+15 downto ppl*w-(i+1)*w);
	end generate; 

end basic;

