-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

-- Possible generic values:
--      HS = {HASH_SIZE_256, HASH_SIZE_512}, 
--		PPL = {2, 4}
-- 		ROMSTYLE = {DISTRIBUTED, COMBINATIONAL}
--
-- Note: ROMSTYLE refers to the type of rom being used in SBOX implementation
--
-- All combinations are allowed

library ieee;
use ieee.std_logic_1164.all;
use work.sha3_pkg.all;
use work.groestl_pkg.all;


entity groestl_top_pq_parallel_ppl is
	generic (
		ROMSTYLE	: integer	:= DISTRIBUTED;
		PPL			: integer 	:= 2;
		HS 			: integer 	:= HASH_SIZE_256);
	port (
		rst 			: in std_logic;
		clk 			: in std_logic;
		src_ready 		: in std_logic_vector(ppl-1 downto 0);
		src_read  		: out std_logic_vector(ppl-1 downto 0);
		dst_ready 		: in std_logic;
		dst_write 		: out std_logic;		
		din				: in std_logic_vector(ppl*w-1 downto 0);
		dout			: out std_logic_vector(w-1 downto 0));
		
end groestl_top_pq_parallel_ppl;


architecture structure of groestl_top_pq_parallel_ppl is
	function get_groest_datasize ( hs : integer ) return integer is
	begin
		if hs = 256 then
			return GROESTL_DATA_SIZE_SMALL;
		elsif hs = 512 then
			return GROESTL_DATA_SIZE_BIG;
		end if;
	end function get_groest_datasize ;

	constant GROESTL_DATA_SIZE : integer := get_groest_datasize( hs );
	
	
	-- fsm1
	signal ein, en_len, en_ctr: std_logic_vector(ppl-1 downto 0);
	signal final_segment : std_logic_vector(ppl-1 downto 0);
	signal c : std_logic_vector(ppl*16-1 downto 0);
	
	-- fsm2
	signal smsg, schain : std_logic_vector(log2(ppl)-1 downto 0);
	signal init1, init2, init3, finalization, wr_state, load_ctr, wr_ctr	:std_logic;
	signal init4 : std_logic;
	signal wr_result : std_logic_vector(ppl-1 downto 0);
	
	-- fsm3		
	signal sout : std_logic_vector(log2(ppl)-1 downto 0);
	signal sid : std_logic;
	signal shift_dout : std_logic_vector(ppl-1 downto 0);
	signal load_id, load_hash : std_logic_vector(ppl-1 downto 0);
begin
   	FsegGen : for i in 0 to ppl-1 generate
		final_segment(i) <= din(w*ppl-i*w-1);	-- MSB of a word
	end generate;


	DatapathInst : entity work.groestl_datapath_pq_parallel_ppl(basic)
	generic map(n=>GROESTL_DATA_SIZE, hs => hs, rom_style=>ROMSTYLE, PPL=>PPL)
	port map (
		clk => clk, rst=>rst, din => din, dout => dout, 		
		c => c, ein => ein, en_ctr => en_ctr, en_len => en_len, smsg => smsg, schain => schain,
		-- fsm2
		init1=>init1, init2=>init2, init3=>init3, init4=>init4, finalization=>finalization, 
		wr_state=>wr_state, wr_result=>wr_result, load_ctr=>load_ctr, wr_ctr=>wr_ctr, 
		-- fsm3
		shift_dout => shift_dout, load_id => load_id, load_hash => load_hash,  sout => sout,  sid => sid				
	);	


	CtrlInst : entity work.groestl_control_pq_parallel_ppl(struct)
    generic map(hs=>hs, n=>GROESTL_DATA_SIZE, PPL=>PPL)
	port map ( 
		clk =>clk, rst=>rst, 
		src_ready => src_ready, src_read => src_read, dst_ready => dst_ready, dst_write => dst_write,	  
		c => c, final_segment => final_segment, ein => ein, en_ctr => en_ctr, en_len => en_len, smsg => smsg, schain => schain,
		-- fsm2
		init1=>init1, init2=>init2, init3=>init3, init4=>init4, finalization=>finalization, 
		wr_state=>wr_state, wr_result=>wr_result, load_ctr=>load_ctr, wr_ctr=>wr_ctr,
		-- fsm3
		shift_dout => shift_dout, load_id => load_id, load_hash => load_hash,  sout => sout,  sid => sid				
	);	
		
end structure;

